//-------newthickness.cpp-------
// This program was written by Paul Dryden, Andy Anderson, Ben Tuttle, and Janna Ballilng
//---------------------------------------------------------------------------
// Thickness.exe reads a cortical .surf file representing a detailed mesh of the cortical surface.
// It takes the vertices defining the mesh elements (from the .hmascii file) and finds the nearest neighbor vertices in
// the cortical .surf file using the program ANN (University of Maryland).  The original cortical
// bone vertices (from the mesh) are used because the geometry is much more representative of 
// the cadaveric pelvis.  Next, the program uses the trabecular .surf file to find the three 
// nearest neighbor vertices to the cortical vertices.  A normalized vector is calculated by taking
// the differences between these cortical verteces and trabecular verteces.  The normals of the 
// cortical file are brought in from the .vtk file (calculated through the use of VTK, see normals.tcl)
// The angle theta between these normals and the normalized vector calculated by this program is then 
// calculated and sorted.  The length of the vector from the cortical to the trabecular points is also 
// calculated.  The angle theta and the length are then each given a weight factor.  In the sorting
// lengths the length based upon the angle is normally taken.  However, if the length based on the angle
// is significantly greater than the length based solely on the length then the program replaces the 
// length based on the angle by the shortest length found.  We found that this method solved most of the
// problems that the program was encountering with sharp edges.  Finally, the distance between the 
// cortical bone and the trabecular bone is reported as the thickness of the bone.  This distance is 
// written out to a new trugrdo (.k) file. Thus a nodal dependent cortical bone thickness 
// is defined for each shell element node.  A log of the calulations is also written to a new (.log) file.
//---------------------------------------------------------------------------

// INCLUDES

#pragma once
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <algorithm>
#pragma hdrstop
#include <ANN\ANNx.h>
#include <vector> 

#include "../findSurfNodes/findMaxNode.cpp"
// GLOBAL VARIABLES

const int NNB = 400; // number of neighbors

// STRUCTURES

struct Pt {
	double vals[3]; 
};

struct IndirectCompare {
	//  Indirect Compare is a structure written by Paul Dryden.  It sorts the values in an array 
	//  from smallest to largest.  It can also be modifies to only give the smallest few (ie 10) 
	//  values
   IndirectCompare(double* values) : m_array(values) {}
   bool operator()(int& m, int& n) {
      return (m_array[m] < m_array[n]);
   }
   double * m_array;
};

struct FileCloser {   //make sure open files get closed
   FileCloser(FILE *f) : m_file(f) {}
   ~FileCloser() { 
	   if(m_file){
		   fclose(m_file);
	   }
   }
   FILE *m_file;
};

struct AnnTableWrap {
  AnnTableWrap(int count) : m_count(count)
  { m_array = annAllocPts(count, 3);   }

  AnnTableWrap() : m_array(NULL), m_count(0) {}

  void Initialize(int count)
  { if(m_array)
      annDeallocPts(m_array);
    m_array = annAllocPts(count, 3);
    m_count = count;
  }

  ~AnnTableWrap() { if(m_array) annDeallocPts(m_array); }

  ANNpointArray m_array;
  int m_count;
  operator ANNpointArray() { return m_array; }
  ANNpointArray get_array() { return m_array; }
  int get_count() { return m_count; }
};

struct norm {
	float x, y, z;
}; 

// USAGE

// This function prints the correct usage of this program to the screen of the user.
void usage()
{
	// mesh input file-			the input .hmascii file of the detailed cortical mesh
	// cortical input file-		the input .surf file of the cortical surface
	// cortical normals file-	the input .vtk  file containing the point normals for all of the nodes in the cortical input file
	// trabecular input file-	the input .surf file of the trabecular surface
	
    printf("Usage: Thickness [mesh input file (.hmascii)] [cortical input file (.surf)] [cortical normals file (.vtk)] [trabecular input file (.surf)] [OPTIONS]\n");
	printf("\nOPTIONS:\n");
	printf("  -ans <outfile (.log)>    = output file name for thickness & angle values; default 'ans.log'\n");
	printf("  -num <shellNum (int)>    = start numbering shells with this number; default '1'\n");
	printf("  -o <outfile (.k)>        = trugrd output file name; default 'out.k'\n");
	printf("  -table <outfile (.log)>  = output file name for all calculation values; default 'table.log'\n");
	printf("  -val <shellVal (int)>    = material value to use for shells; defalut '1'\n");
	exit(-1);
}

// FUNCTIONS

// This function takes the dotproduct of vectors a and b.
inline double dotProduct(ANNpoint a, ANNpoint b) {
	return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] ;
}

// This function takes the length of a vector by taking the square root of the sum of all parts squared
inline double LenVector(ANNpoint a) {
   return sqrt(dotProduct(a,a));
}

// This function normalizes the theta and length values to a scale from 1-10
// Arguement A is the theta or length variable and B is the max value expected in the A array
// Example theta has values from 0 to 130 A=theta B=130
inline double NormalVal(double a, double b) {
	return 1+((a)*((10-1)/(b-0)));
}

// These functions use 2 weighting scales.  This is to solve the problems that the program encountered when
// dealing with sharp edges.  
//   WeightFactor1 weighting based on length
//   WeightFactor2 weighting based on angle
inline double WeightFactor1(double a, double b) {
	return  (a*0 + b*1);
}
inline double WeightFactor2(double a, double b)  {
	return  (a*1 + b*0);
}

// This function normalizes the vector you get by subtracting coordinates.  It divides 
// the component of the vector by the total length of the vector
void NormalizeVector(ANNpoint a) {
	double len = LenVector(a);
    a[0] /= len;
	a[1] /= len;
	a[2] /= len;
}

// This function does an indirect sort
void IndirectSort(double* values, int* indices, int n) {
  std::sort(indices, indices+n, IndirectCompare(values));
}

// This function calculates the distance from the cortical surface (query point) to the trabecular surface
double lenPtToSurf(ANNpoint pt, ANNpoint normals, ANNpoint surfPts, int* pInt, int count, double& mintheta)
{
	double length[NNB], theta[NNB], dist;
    int indices[NNB];  // indices to the trabecular surface array.
    int  i;

	for (i=0;i<NNB; i++){
        double diff[3];
		ANNpoint neighbor = &surfPts[3*pInt[i]];  // gets surfPts to return three values like the other ANNpoints
		indices[i] = i;

		// Get the vector from the cortical to the trabecular surface
		diff[0] = pt[0]-neighbor[0];
		diff[1] = pt[1]-neighbor[1];
		diff[2] = pt[2]-neighbor[2];
		Pt *normal = (Pt *) normals;

		// Calculate the length of the vector from the cortical to the trabecular surface
		length[i] = LenVector(diff);
		// Normalize the calculated vector 
		NormalizeVector(diff);
		// Check to see that the vector is normalized  ans~1
 //		assert( fabs(1-LenVector(normals))<1e-4 );
		// Calculate the dot product of the surface normal of the qp and the normalized 
		// vector calculated from the qp to the trabecular surface
		double dp = dotProduct(diff, normals);
		double thetacos=acos(dp);
		// Calculate the angle between the two vectors
		theta[i] = fabs(acos(dp)*(180/3.145));
	}
    // Sort the one hundred neighbors by their theta
 // std::sort(&indices[0], &indices[NNB-1], IndirectCompare(theta));
 // shortest theta is theta[indices[0]];

	// Apply weight factors based on the theta and length of the vectors to determine the 
	// nearest but most accurate neighbor
    dist=length[indices[0]];
	mintheta=(theta[indices[0]]);
	double Normtheta=NormalVal(theta[indices[0]],110);
	double Normlength=NormalVal(length[indices[0]],10);
//double weighted=weightFactor(theta[indices[0]]), length[indices[0]]);
	double weighted1=WeightFactor1(Normtheta, Normlength);
	double weighted2=WeightFactor2(Normtheta, Normlength);	
	double distL=length[indices[0]];
	double distT=length[indices[0]];
	double minthetaL=theta[indices[0]];
	double minthetaT=theta[indices[0]];
	for (i=1; i<NNB; i++){
        double test1=NormalVal(theta[indices[i]],110);
		double test2=NormalVal(length[indices[i]],10);
		double test3 = WeightFactor1(test1, test2);
		double test4 = WeightFactor2(test1, test2);
		double dist1= length[indices[i]];
		double dist2= length[indices[i]];
		double mintheta1= (theta[indices[i]]);
		double mintheta2= (theta[indices[i]]);
	//double test3 = weightFactor(theta[indices[i]],length[indices[i]]);
	//  double test = weightFactor(ang[0], distance[0]);

		if (test3< weighted1){
			weighted1=test3;
			distL=dist1;
			minthetaL = mintheta1;
		}
		if (test4<weighted2) {
			weighted2=test4;
			distT= dist2;
			minthetaT= mintheta2;
		}
	}

	if (distT>distL*1.5){
		distT=distL;
		minthetaT=minthetaL;
	}
	if (distT<dist){
		dist=distT;
		mintheta=minthetaT;
	}
 //if (dist>4){
//		dist=4;
 //}	
	
	return dist;
	// returns the distance between the query point and the nearest neighbor for 
    // output to the trugrdo file
}

// This function attempts to open the file (path) for either reading(rt) or writing(wb) (see fopen).
// The opened file pointer is returned else error.
FILE * fail_or_open(const char *path, const char *opt)
{
   FILE *f = fopen(path, opt);
   if(!f) {
     printf("Failure opening file %s\n", path);
     exit(-13);
   }
   return f;
}
// This function will parse the vertices values from the .k mesh file.  The number of vertices 
// found is saved as vertices and the values of each are saved in the Table.  The status
// of the program is returned (0 on success else an integer <0).
int parse_mesh_file(char *fmeshName, AnnTableWrap& Table, int& vertices)
{
	FILE *fmesh = fail_or_open(fmeshName, "rt");
	FILE *fmeshTemp = fail_or_open("pelvisfinal_med.d", "rt");
	if(fmesh==NULL||fmeshTemp==NULL){
		printf("Failed to open .k file\n");
		return -1;
	}

	// finding maximum value of Vertices from shells in .k file
	vertices=0;
	vertices=findMaxNode(fmesh,false);
	if(vertices==0){
		printf("Invalid number of vertices found in .k file\n");
		return -2;
	}


	// finding verticies in file
	char line[1024];
	
	if(NULL==fgets(line, sizeof(line), fmeshTemp) ){
		printf("Failed to read first line of .k file\n");
		return -3;
	}

	// saving values of vertices to querypointTable
	Table.Initialize(vertices);
	bool doPrint=false;

	for(int i=0;i<vertices;i++){
		ANNpoint pt = Table.get_array()[i];
		float temp[3];
		if(3!=sscanf(line,"%*d%*d%e%e%e",&temp[0],&temp[1],&temp[2]))
		{
			printf("Couldn't read value of vertex #%d in .k file\n",i+1);	
			return -4;
		}
		pt[0]=temp[0];
		pt[1]=temp[1];
		pt[2]=temp[2];
		if(doPrint){
			printf("v#%d : %e %e %e\n",i+1,pt[0],pt[1],pt[2]);
		}
		if(NULL == fgets(line, sizeof(line), fmeshTemp)){
			printf("Reached end of .hmascii file before finding all vertices\n");
			return -5;
		}
	}

	return 0;
}

// This function will parse the vertices values from the fsurf file.  The number of vertices 
// found is saved as vertices and the values of each are saved in the Table.  The status
// of the program is returned (0 on success else an integer <0).
int parse_off_file(FILE *foff, AnnTableWrap& Table, int& vertices)
{
	if(foff==NULL){
		printf("Failed to open .off file\n");
		return -1;
	}
	char line[1024];
	vertices = 0;

	// finding number of Vertices from .off file
	if(NULL==fgets(line, sizeof(line), foff) ){
		printf("Failed to read first line of .off file\n");
		return -2;
	}
	while (NULL==strstr(line,"NOFF")){
		if(NULL == fgets(line, sizeof(line), foff)){
			printf("Reached end of .off file before finding 'NOFF'\n");
			return -3;
		}
	}
	if(NULL == fgets(line, sizeof(line), foff)){
			printf("Expected number of verticies, elements, & lines in .off'\n");
			return -4;
		}
	if(1!=sscanf(line,"%d",&vertices)){
		printf("Couldn't read number of vertices in .off file\n");	
		return -5;
	}
	if(vertices==0){
		printf("Invalid number of vertices found in .off file\n");
		return -6;
	}

	// saving values of vertices to querypointTable
	Table.Initialize(vertices);
	for(int i=0;i<vertices;i++){
		if(NULL == fgets(line, sizeof(line), foff)){
			printf("Reached end of .off file before finding all vertices\n");
			return -7;
		}
		ANNpoint pt = Table.get_array()[i];
		if(3!=sscanf(line,"%lg %lg %lg", &pt[0], &pt[1], &pt[2])){
			printf("Couldn't read value of vertex #%d in .off file\n",i+1);	
			return -8;
		}
		if(i==0){
			printf("off verticie 1 : %lf %lf %lf\n",pt[0],pt[1],pt[2]);
		}else if(i==(vertices-1)){
			printf("off verticie last : %lf %lf %lf\n",pt[0],pt[1],pt[2]);
		}	
	}

	return 0;
}

// This function will parse the normal values from the fnorm (.vtk) file. The normal values
// are saved into the array normals.  The status of the program is returned (0 on success 
// else an integer <0). 
int parse_off_for_norm(FILE *foff, int expected, ANNpoint* dest) { 
	if(foff==NULL){
		printf("Failed to open .off file\n");
		return -1;
	}

	*dest = (ANNpoint ) (new double[3*expected]); // allocate memory to hold the normal array
    double *fn = (double*) *dest;
	char line[1000];

	// finding 'OFF' in .off file
	if(NULL==fgets(line, sizeof(line), foff) ){
		printf("Failed to read first line of .off file\n");
		return -2;
	}
	while (NULL==strstr(line,"NOFF")){
		if(NULL == fgets(line, sizeof(line), foff)){
			printf("Reached end of .off file before finding 'NOFF'\n");
			return -3;
		}
	}
	if(NULL == fgets(line, sizeof(line), foff)){
			printf("Expected number of verticies, elements, & lines in .off'\n");
			return -4;
		}

	// finding and saving normal values from .vtk file
	for(int i=0; i<expected; i++){
		if(NULL == fgets(line, sizeof(line), foff)){
			printf("Reached end of .vtk file before finding all normals: %d/%d\n",i+1,expected);
			return -6;
		}
		if(3!=sscanf(line,"%*lf %*lf %*lf %lf %lf %lf",fn, fn+1, fn+2)){
			printf("Couldn't read value of normals #%d in .off file\n",i+1);	
			return -5;
		}
		if(i==0){
			printf("off normal 1 : %lf %lf %lf\n",fn[0],fn[1],fn[2]);
		}		
		fn +=3;
	}
	fn-=3;
	printf("off normal last : %lf %lf %lf\n",fn[0],fn[1],fn[2]);
	
	return 0; 
} 

// Calculates values for thickness and angle
// and prints to .log files
int calculate_angles_thickness(FILE *flog, FILE *flogshort, 
						AnnTableWrap& meshTable, AnnTableWrap& cortTable, AnnTableWrap& trabTable,
						ANNpoint normals,
						float *thickness)
{
	int vertices=meshTable.get_count();
	ANNkd_tree kdCort(cortTable.get_array(), cortTable.get_count(), 3);
	ANNkd_tree kdTrab(trabTable.get_array(), trabTable.get_count(), 3);


	int icort;		//index to point in cortTable
	int itrab[NNB*3];	//indecies to points neighboring icort
	double dnb[NNB];//distances between neighbor points and icort
	double ans1; //storage for 1st calculation of distances between icort and the trabecular surface
	double mintheta=0;	//storage for final calculation.
	
	
	fprintf(flog,"MeshNode\tCorticalNode\tX\t\tY\t\tZ\t\tThickness\tAngle\n");
	fprintf(flogshort, "MeshVertex\tThickness\tAngle\n");
	for(int k=0;k<vertices;k++){
		thickness[k]=0.0f;	//storage for final calculation and adjustment of thickness
		
		//find point nearest to meshTable point in cortTable (call it icort)
		kdCort.annkSearch(      //aprox k near neighbor search
			meshTable.get_array()[k],		// query point on mesh
            1,		// number of near neighbors to return
            &icort,	// where to save index to nearest neighbor (cortical)
            dnb,		// where to save distance to near neighbors
            0.0);    // eps
		ANNpoint cort = cortTable.get_array()[icort];
		//find 3 closest points to cortTable point (cort) in trabTable (call them itrab[])
		kdTrab.annkSearch(		// approx k near neighbor search
            cort,	// query point on cortical surface
            NNB,		// number of near neighbors to return
            itrab,	// where to save indecies to neighbors (trabecular)
            dnb,		// where to save distance to near neighbors
            0.0);    // eps
		//call lenPtToSurf function to get the dist between the cortical point (cort) and the trabecular neighbors (itrab[])
		ans1 = lenPtToSurf(
			(*cortTable.get_array())+3*icort,
			(ANNpoint)(&(normals[3*icort])),
            (*trabTable.get_array()),
			itrab,
            NNB,
			mintheta);

		thickness[k] = ans1;
//Changes for STATS DETERMINATIONS
		//adjust answer from previous step
		//thickness[k] = ans1 -.62222;
		/*thickness[k]-=0.44f;
		if(thickness[k]<0.44)
			thickness[k]=0.44f;
		else if(thickness[k]>4.0)
			thickness[k]=4.0f;*/
	
		//print values to table.log file
		fprintf(flog, "%8d\t%8d\t\t%0.3e\t%0.3e\t%0.3e\t%0.3e\t%2.2f\n",
				k+1,icort+1,cort[0],cort[1],cort[2],thickness[k],mintheta);
		fprintf(flogshort, "%d\t\t%0.3e\t%2.2f\n",k+1,thickness[k],mintheta);
	}
	
	return 0;
}
	
// This function will parse the triangle values from the fmesh file.  It will then find the matching
// value in the meshTable.  For each value, the closest cortTable value is calculated.  The thickness
// and angles to the trabecular surface are then found (using the trabTable values).  Results are 
// printed to the foutk, flog, and flogshort files.
int calculate_and_print(FILE *fmesh, FILE *foutk,
						float *thickness, 
						int newNum, int newVal)
{
	// Printing to .k  file using triangles from off file
	char line[1024];
	int triNumber=0,triValues[3]={0};
	int triangles=0;

	fprintf(foutk,"*ELEMENT_SHELL_THICKNESS\n");
	fprintf(foutk,"$\n$ ELEMENT CARDS FOR SHELL ELEMENTS\n$\n");
	while(NULL != fgets(line, sizeof(line), fmesh)){
		triNumber++;
		if(3!=sscanf(line,"%*d%d%d%d",
			&triValues[0],&triValues[1],&triValues[2]))
		{
			printf("Couldn't read value of triangle #%d in .k file\n",triNumber+1);
			return -4; 
		}
		if(triNumber==1){
			printf("off triangle 1 : %d %d %d\n",triValues[0],triValues[1],triValues[2]);
		}

		//print triangle#, surface material, and triangle values to .k file (last value is repeated)
		fprintf(foutk, "%8d%8d%8d%8d%8d%8d\n",newNum,newVal,triValues[0]+1,triValues[1]+1,triValues[2]+1,triValues[2]+1);
		
		//print thickness for each triangle value to .k file (last value repeated)
		fprintf(foutk,"%16.8f%16.8f%16.8f%16.8f\n",
			thickness[triValues[0]], thickness[triValues[1]], thickness[triValues[2]], thickness[triValues[2]] );
	
		newNum++;
	}//end while for triangles

	printf("off triangle last: %d %d %d\n",triValues[0],triValues[1],triValues[2]);
			
	fprintf(foutk,"*END\n");
	printf("Found %d triangles in .hmascii file\n",triNumber);
 
	return 0;
}

int main(int argc, char* argv[])
{
	printf("\n");
	if(argc<2) {
		usage();
		return -1;
	}

	FILE *fmesh, *fcort, *fnorm, *ftrab, *fk, *ftable, *fans;
	char *fmeshName=argv[1];
	fmesh = fail_or_open(argv[1], "rt");
	fcort = fail_or_open(argv[1], "rt");
	fnorm = fail_or_open(argv[1], "rt");
	ftrab = fail_or_open(argv[2], "rt");
	FileCloser fc1(fmesh), fc2(fcort), fc3(fnorm), fc4(ftrab);

	int shellNum=1,shellVal=1;
	bool nokFile=true,notableFile=true,noansFile=true;
	char toConvert[20];
	for(int i=3;i<argc;i++){
		if(0==strcmp(argv[i],"-ans")){
			fans = fail_or_open(argv[i+1], "wb");
			printf("writing file : %s\n",argv[i+1]);
			noansFile=false;
			i++;
		}else if(0==strcmp(argv[i],"-num")){
			strcpy(toConvert,argv[i+1]);
			if(1!=sscanf(toConvert,"%d",&shellNum)){
				printf("entered invalid number for initial shell index\n");
				usage();
				return -2;
			}
			else{
				i++;
			}
		}else if(0==strcmp(argv[i],"-o")){
			fk = fail_or_open(argv[i+1], "wb");
			printf("writing file : %s\n",argv[i+1]);
			nokFile=false;
			i++;
		}else if(0==strcmp(argv[i],"-table")){
			ftable = fail_or_open(argv[i+1], "wb");
			printf("writing file : %s\n",argv[i+1]);
			notableFile=false;
			i++;
		}else if(0==strcmp(argv[i],"-val")){
			strcpy(toConvert,argv[i+1]);
			if(1!=sscanf(toConvert,"%d",&shellVal)){
				printf("entered invalid number for new shell value\n");
				usage();
				return -2;
			}
			else{
				i++;
			}
		}
	}
	if(noansFile){
		fans = fail_or_open("ans.log", "wb");
		printf("writing file : ans.log\n");
	}
	if(nokFile){
		fk = fail_or_open("out.k", "wb");
		printf("writing file : out.k\n");
	}if(notableFile){
		ftable = fail_or_open("table.log", "wb");
		printf("writing file : table.log\n");
	}
	FileCloser fc5(fk), fc6(ftable), fc7(fans);

	printf("start shells with : %d\n",shellNum);
	printf("assinging shell value : %d\n",shellVal);
	printf("using mesh file       : %s\n",argv[1]);
	printf("using cortical file   : %s\n",argv[1]);
	printf("using normals file    : %s\n",argv[1]);
	printf("using trabecular file : %s\n\n",argv[2]);



	int status;

  // Finding vertex values from .hmascii and .surf files
  AnnTableWrap meshTable, cortTable, trabTable;
  int meshVertices, cortVertices, trabVertices;

  status = parse_off_file(fmesh, meshTable, meshVertices);
  if(status!=0) {
    printf("Failure parsing vertices in mesh.hmascii file\n");
    return -1;
  }
  else
	  printf("Used %d vertices in mesh surface .hmascii file\n",meshVertices);
  status = parse_off_file(fcort, cortTable, cortVertices);
  if(status!=0) {
    printf("Failure parsing vertices in cort.surf file\n");
    return -2;
  }
  else
    printf("Found %d vertices in cort.surf file\n",cortVertices);
  status = parse_off_file(ftrab, trabTable, trabVertices);
  if(status!=0) {
    printf("Failure parsing vertices in trab.surf file\n");
    return -3;
  }
  else
    printf("Found %d vertices in trab.surf file\n",trabVertices);

  // Finding normal values from .vtk file
  ANNpoint normals; // pointer to array of normals
  fclose(fcort);
  fcort = fail_or_open(argv[1], "rt");
  status = parse_off_for_norm(fcort, cortVertices, &normals); 
  if(status!=0) {
    printf("Failure parsing normals in norm.vtk file\n");
    return -4;
  }
  else
    printf("Found %d normals in norm.vtk file\n",cortVertices);
 
  //Calculating the thickness and angles between cortical and trabecular surfaces and
  //printing results to .out files
  float *tvalues = new float[meshVertices];
  
  status=calculate_angles_thickness(ftable,fans,meshTable,cortTable,trabTable,normals,tvalues);
  if(status!=0) {
    printf("Failure calculating values\n");
    return -5;
  }
  else
    printf("Finished calculating values\n");


  //Finding triangle values from .ascii file and
  //printing results to .k file
  status=calculate_and_print(fmesh, fk, tvalues, shellNum, shellVal );
  if(status!=0) {
    printf("Failure parsing triangles in .ascii file\n");
    return -6;
  }
  else
    printf("Finished parsing triangles in .ascii file\n");

  //cleanup
  delete [] normals;
  delete [] tvalues;

  return 0;
}
