// Mesh.cpp: implementation of the CMesh class.

#include "stdafx.h"

#pragma once
#include "Mesh.h"

CMesh::CMesh(int elements, int nodes, int surfs, int minIntensity, int maxIntensity)
{
	m_nElems = elements;
	m_nNodes = nodes;
	m_nSurfs = surfs;
	minINTENSITY = minIntensity;
	maxINTENSITY = maxIntensity;
	minELASTICMODULUS = intensityToEM(minINTENSITY);
	maxELASTICMODULUS = intensityToEM(maxINTENSITY);
	m_pNode = new float[3*m_nNodes];
	m_pEM = new double[m_nNodes];
	m_eEM = new double[m_nElems];
	m_eVal = new int[m_nElems];
}

CMesh::~CMesh()
{	
	m_nElems = 0;
	m_nNodes = 0;
	m_nSurfs = 0;
	minINTENSITY = 0;
	maxINTENSITY = 0;
	minELASTICMODULUS = 0.0;
	maxELASTICMODULUS = 0.0;
	delete [] m_pNode; m_pNode = NULL;
	delete [] m_pEM; m_pEM = NULL;
	delete [] m_eEM; m_eEM = NULL;
	delete [] m_eVal; m_eVal = NULL;
}

bool CMesh::ReadMesh(const char* szFileName)
{
	FILE* fp = fopen(szFileName, "rt");
	if (fp == NULL)
		return false;

	char line[1000];

	// find *NODE in file
	if(NULL==fgets(line, sizeof(line), fp) ){
		printf("Failed to read first line of nodes (.k) file\n");
		return false;
	}
	while (NULL==strstr(line,"*NODE")){
		if(NULL == fgets(line, sizeof(line), fp)){
			printf("Reached end of nodes (.k) file before finding '*NODE'\n");
			return false;
		}
	}

	// save nodes in m_pNode array
	float* pn = m_pNode;

	for (int i=0; i<m_nNodes; i++, pn += 3)
	{
		fgets(line, sizeof(line), fp);
		sscanf(line, "%*d%*c%g%*c%g%*c%g\n", pn, pn+1, pn+2); 
	}

	fclose(fp);
	return true;
}

bool CMesh::findModulus(CImageSet& set, double xadj, double yadj, double zadj)
{
	FILE* flog = fopen("findModulus.log","wt");
	if (flog==NULL)
		return false;

	float* pn = m_pNode;
	int i, j, k, kadjusted;
	double minpEM=maxELASTICMODULUS,maxpEM=0,avepEM=0;
	int minIn=maxINTENSITY,maxIn=0,aveIn=0,count=0;
	int belowMinIn=0,aboveMaxIn=0;
	
	const int neighbors = 9;
	int In[neighbors];
	int AvInten=0,numToAve=0;

	fprintf(flog,"\n");
	fprintf(flog,"\tNode     i       j       k\t In\taveInten	EM\n");
	//calculate Elastic Modulus for each node
	for (int m=0; m<m_nNodes; m++, pn+=3)
	{ 
		//get volume coordinate(pixels) from node coordinate(mm) values
		i = (int) (pn[0] * xadj);
		j = (int) (pn[1] * yadj); 
		k = (int) (pn[2] * zadj);
		kadjusted = k;  //z axis was inverted from the amira orientation, so adjusted by subtracting from # of slices(m_count)
		//kadjusted = (set.m_count-1)-k;  //z axis was inverted from the amira orientation, so adjusted by subtracting from # of slices(m_count)

		//if on an edge, move it over so nearest neighbors will exist.
		if(i<1)		 i=1;
		if(i>set.m_cx-2) i=set.m_cx-2;
		if(j<1)	 j=1;
		if(j>set.m_cy-2) j=set.m_cy-2;
		if(kadjusted<0)		 kadjusted=0;
		if(kadjusted>set.m_cz*set.m_count-1) kadjusted=set.m_cz*set.m_count-1;

		////find intensity of self and 8 nearest neighbors from CT values
 		In[0] = *set.Pixel(i-1, j-1, kadjusted);
		In[1] = *set.Pixel(i, j-1, kadjusted);
		In[2] = *set.Pixel(i+1, j-1, kadjusted);
		In[3] = *set.Pixel(i-1, j, kadjusted);
		In[4] = *set.Pixel(i, j, kadjusted);
		In[5] = *set.Pixel(i+1, j, kadjusted);
		In[6] = *set.Pixel(i-1, j+1, kadjusted);
		In[7] = *set.Pixel(i, j+1, kadjusted);
		In[8] = *set.Pixel(i+1, j+1, kadjusted);
		
		//AvInten = Average Pixel Intensity
		AvInten=0;
		numToAve=0;
		for(int inIndex=0;inIndex<9;inIndex++){
			if(In[inIndex]<minINTENSITY){}
			else if(In[inIndex]>maxINTENSITY){}
			else{
				AvInten+=In[inIndex];
				numToAve++;
			}
		}
		if(numToAve!=0){
			AvInten/=numToAve;
		}else{
			AvInten=In[4];
		}

		//m_pEM = Elastic Modulus for the node
		if(m<m_nSurfs){
			//if node is on the surface, make EM=0.0
			m_pEM[m] = 0.0;
		}else if(AvInten<minINTENSITY){
			//set EM of minimum AvInten different from surface nodes 
			m_pEM[m] = minELASTICMODULUS;
			belowMinIn++;
		}else if(AvInten>maxINTENSITY){
			//set EM of maximum AvInten
			m_pEM[m] = maxELASTICMODULUS;
			aboveMaxIn++;
		}else{
            // Calculate the Elastic Modulus
			m_pEM[m] = intensityToEM(AvInten);
				
			//figure statistics for nodes with calculate EMs
			if(m_pEM[m]<minpEM) minpEM = m_pEM[m];
			if(m_pEM[m]>maxpEM) maxpEM = m_pEM[m];
			avepEM+=m_pEM[m];
			if(AvInten<minIn) minIn=AvInten;
			if(AvInten>maxIn) maxIn=AvInten;
			aveIn+=AvInten;
			count+=1;
		}
		fprintf(flog,"%8d%8d%8d%8d%8d%8d\t%13f\n",m+1,i,j,k,In[4],AvInten,m_pEM[m]);
	}
	//print statistics to log
	if(count!=0){
		avepEM=avepEM/count;
		aveIn=aveIn/count;
	}
	fprintf(flog,"\n NODE STATISTICS \n");
	fprintf(flog,"\n\nMinEM:\t%13.6f\nMaxEM:\t%13.6f\nAveEM:\t%13.6f\n",minpEM,maxpEM,avepEM);
	fprintf(flog,"\nMinIn:\t%8d\nMaxIn:\t%8d\nAveIn:\t%8d\n",minIn,maxIn,aveIn);
	fprintf(flog,"\nSurface Nodes:\t%8d",m_nSurfs);
	fprintf(flog,"\nBelow %d In:\t%8d",minINTENSITY,belowMinIn);
	fprintf(flog,"\nAbove %d In:\t%8d",maxINTENSITY,aboveMaxIn);
	
	fclose(flog);
	return true;
}

bool CMesh::labelElements(const char* meshFileName, int startMaterial)
{
	FILE* flog = fopen("labelElements.log","wt");
	if (flog==NULL)
		return false;
	
	FILE* fp = fopen(meshFileName, "rt");
	if (fp == NULL)
		return false;

	char line[1000];

	// find *ELEMENT_SOLID in file
	if(NULL==fgets(line, sizeof(line), fp) ){
		printf("Failed to read first line of mesh (.k) file\n");
		return false;
	}
	while (NULL==strstr(line,"*ELEMENT_SOLID")){
		if(NULL == fgets(line, sizeof(line), fp)){
			printf("Reached end of mesh (.k) file before finding '*ELEMENT_SOLID'\n");
			return false;
		}
	}
	while(NULL!=strstr(line,"*")||NULL!=strstr(line,"$")){
		if(NULL == fgets(line, sizeof(line), fp)){
			printf("Reached end of mesh (.k) file before finding elements\n");
			return false;
		}
	}

	//find average EM for the 4 element verticies
	int findNode[4]={0};
	double aveMin=maxELASTICMODULUS,aveMax=0,aveAve=0;
	int count=0;

	fprintf(flog,"\n");
	fprintf(flog,"Element\t	EM     EMnode1      EMnode2      EMnode3      EMnode4\n");
	for(int i=0;i<m_nElems;i++){
		if(4!=sscanf(line,"%*d%*c%*d%*c%d%*c%d%*c%d%*c%d",&findNode[0],&findNode[1],&findNode[2],&findNode[3])){
			printf("Error scanning elements\n");
			return false;
		}

		m_eEM[i]=(m_pEM[findNode[0]-1] + m_pEM[findNode[1]-1] + m_pEM[findNode[2]-1] + m_pEM[findNode[3]-1])/4.0;
        if(m_eEM[i]<aveMin) aveMin=m_eEM[i];
		if(m_eEM[i]>aveMax) aveMax=m_eEM[i];
		aveAve+=m_eEM[i];
		count++;

		fprintf(flog,"%8d%13f%13f%13f%13f%13f\n",i+1,m_eEM[i],m_pEM[findNode[0]-1],m_pEM[findNode[1]-1],m_pEM[findNode[2]-1],m_pEM[findNode[3]-1]);
	
		//get next line of shell info
		fgets(line, sizeof(line), fp);
	}
	aveAve/=count;
	fprintf(flog,"\n ELEMENT STATISTICS \n");
	fprintf(flog,"MinEM:\t%13.6f\nMaxEM:\t%13.6f\nAveEM:\t%13.6f\n",aveMin,aveMax,aveAve);
	
	fclose(fp);

	//Asign part number (bin) according to average EM.
	double EMrange=(aveMax-aveMin);
	string temp;
	bool evenDivision=true;
	double evenBinRange=0;
	int numberOfBins=21;
	int* numberInBin=new int[numberOfBins]; 
	double* binMax=new double[numberOfBins]; 
	bool FoundRange=false;
	int binVal= startMaterial;
	int binNum=0;
	
	while(true) //this loop exits at the break when user is satisfied with bin distribution
	{
		//set count in each bin to 0
		for(int m=0;m<numberOfBins;m++){
			numberInBin[m]=0;
		}

		//asign bins with even division
		if(evenDivision)
		{
			//set binMax to fraction of total range for each bin
			evenBinRange = EMrange/(numberOfBins-2)+0.000001;
			binMax[0]=0;
			for(int n=1;n<numberOfBins-1;n++){
                binMax[n]=evenBinRange*n;
			}
			binMax[numberOfBins-1]=maxELASTICMODULUS;
		}
		//ssign bins with user given max for each bin
		else
		{
			printf("Enter max number for elastic modulus value in bin #: (double format)\n");
			//set binMax from user for each bin
			binMax[0]=0;
			for(int p=1;p<numberOfBins-1;p++){
				printf("  %d - ",p);
				scanf("%lf",&binMax[p]);
			}
			binMax[numberOfBins-1]=maxELASTICMODULUS;
			printf("\n\n");
		}
        
		//asign each element to a bin
		for(int j=0;j<m_nElems;j++){
				FoundRange=false;
			if(m_eEM[j]==0.0){
				//elements all on surface
				m_eVal[j]=startMaterial;
				numberInBin[0]++;
				FoundRange=true;
			}else if(m_eEM[j]>=maxELASTICMODULUS){
				//elements all with maxEM
				m_eVal[j]=startMaterial+numberOfBins-1;
				numberInBin[numberOfBins-1]++;
				FoundRange=true;
			}else{
				//find bin and asign
				binVal=startMaterial;
				binNum=0;
				while(!FoundRange){
					binVal++;
					binNum++;
					if(m_eEM[j]<binMax[binNum]){
						m_eVal[j]=binVal;
						numberInBin[binNum]++;
						FoundRange=true;
					}
				}
			}
		}

		//print bin distributions to screen for user to evaluate
		printf("\nBin Distributions:\n\n");
		printf("Bin#\t\t\tRange\t\t\t#in Range\n");
		printf("%5d %13.6f - %13.6f %d\t all on surface\n",0,0.0,binMax[0],numberInBin[0]);
		for(int q=1;q<numberOfBins;q++){
			printf("%5d %13.6f - %13.6f %d\n",q,binMax[q-1],binMax[q],numberInBin[q]);
		}
		printf("\n");

		//ask user if satisfied and set values to repeat
		printf("Are you satisfied with this bin distribution? (y or n)\n");
		
		cin>>temp;
		if(temp=="y")
			break;
		printf("Would you like to use even distribution or give values for the ranges? (e or v)\n");
		cin>>temp;
		if(temp=="e")
			evenDivision=true;
		else
			evenDivision=false;
		printf("How many bins would you like to make? (integer format, at least 2)\n");
		cin>>numberOfBins;


		//clean up for next itteration	
		delete [] numberInBin;
		numberInBin=new int[numberOfBins]; 
		delete [] binMax;
		binMax=new double[numberOfBins]; 
	
	}
	
	//print final bin distribution to file for future use
	fprintf(flog,"\nMATERIAL DISTRIBUTION\n");
	fprintf(flog,"Bin#\t\t\tRange\t\t\t#in Range\n");
	fprintf(flog,"%5d %13.6f - %13.6f %d\t all on surface\n",0,0.0,binMax[0],numberInBin[0]);
	for(int r=1;r<numberOfBins;r++){
		fprintf(flog,"%5d %13.6f - %13.6f %d\n",r,binMax[r-1],binMax[r],numberInBin[r]);
	}

	//clean up 
	delete [] numberInBin;
	delete [] binMax;
	
	fclose(flog);
	return true;
}

bool CMesh::printElements(const char* meshFileName, const char* newFileName, bool doCopy, int startElem)
{
	FILE* fmesh = fopen(meshFileName, "rt");
	FILE* fout = fopen(newFileName, "wt");
	if (fmesh==NULL || fout==NULL)
		return false;

	char line[1000];

	// find *ELEMENT_SOLID in file
	if(NULL==fgets(line, sizeof(line), fmesh) ){
		printf("Failed to read first line of mesh (.k) file\n");
		return false;
	}
	while (NULL==strstr(line,"*ELEMENT_SOLID")){
		if(doCopy){
			fputs(line,fout);
		}
		if(NULL == fgets(line, sizeof(line), fmesh)){
			printf("Reached end of mesh (.k) file before finding '*ELEMENT_SHELL'\n");
			return false;
		} 
	}
	while(NULL!=strstr(line,"*")||NULL!=strstr(line,"$")){
		 if(doCopy){
			fputs(line,fout);
		 }
		 if(NULL == fgets(line, sizeof(line), fmesh)){
			printf("Reached end of mesh (.k) file before finding shells\n");
			return false;
		}
	}

	//print elements 
	int printNode[4];
	
	for(int i=0;i<m_nElems;i++){
		if(4!=sscanf(line,"%*d%*c%*d%*c%d%*c%d%*c%d%*c%d",&printNode[0],&printNode[1],&printNode[2],&printNode[3])){
			printf("Error scanning elements\n");
			return false;
		}
		fprintf(fout,"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",startElem+i,m_eVal[i],printNode[0],printNode[1],printNode[2],printNode[3],printNode[3],printNode[3],printNode[3],printNode[3]);
		
		//get next line of element info
		fgets(line, sizeof(line), fmesh);
	}
	if(doCopy){
		fputs(line,fout);
		while(NULL != fgets(line, sizeof(line), fmesh)){
			fputs(line,fout);
		}
	}

	fclose(fmesh);
	fclose(fout);
	return true;
}
