// sphereFit.cpp : Fits a Sphere to given data points using Least Squares Approximation.

// Written by Janna Balling
// 6/04/04
// Alogrithms and loop execution found at 
// http://www.magic-software.com/Documentaion/LeastSquaresFitting.pdf  #6 pg.5
// http://www.magic-software.com/Source/Approximation/WmlApprSphereFit3.cpp

// The function sphereFit finds a sphere origin (a,b,c) and radius that best 
// fits (using Least Squares Approximation) a list of given data points 
// (datanum, data[x,y,z]).  It starts fitting with an initial guess (avex,avey,avez) 
// and repeats untill the change in a, b, or c is neglegible or the given max 
// number of itterations (maxi) is reached, which ever comes first.

// returns true if didn't reached maxi.  False if reached maxi.

// The Least Squares Solution equations used are:
// radius:
// radius = sum(Li)/datanum
// a = sum(xi)/datanum + sum(Li)sum((ap-xi)/Li)/(datanum*datanum)
// b = sum(yi)/datanum + sum(Li)sum((bp-yi)/Li)/(datanum*datanum)
// c = sum(zi)/datanum + sum(Li)sum((cp-zi)/Li)/(datanum*datanum)
//   where
// sum is the summation for i from 1 to datanum
// Li=sqrt( (xi-ap)^2 + (yi-bp)^2 + (zi-cp)^2 )
// x,y,z are the data points
// ap,bp,cp are the previous calculations for a,b,c

#include <math.h>  //for sqrt and fabs functions
#include <float.h> //for DBL_EPSILON (really small double ~0)

bool sphereMapper::sphereFit (int maxi)
{
	double xave=0,yave=0,zave=0;

	// compute sum(datai)/datanum since these terms don't vary with itterations
	for(int i=0;i<datanum;i++){
		xave+=datax[i];
		yave+=datay[i];
		zave+=dataz[i];
	}
	xave/=datanum;
	yave/=datanum;
	zave/=datanum;

	double ap,bp,cp;
	if(a==0.0&&b==0.0&&c==0.0){
        ap=xave,bp=yave,cp=zave;
	}else{
		ap=a,bp=b,cp=c;
	}
    printf(" Intitial origin guess: %lf %lf %lf\n",ap,bp,cp);
	radius=0,a=0,b=0,c=0;

	double tolerance=0.00000001;

	// compute a,b,c & radius for each itteration
	int j=0;
	for(j=0;j<maxi;j++){
		double Li=0;
		double diffa=0,diffb=0,diffc=0;
		double sumda=0,sumdb=0,sumdc=0;
		double sumLi=0;

		// compute sums for itteration dependant terms
		for(int n=0;n<datanum;n++){
			diffa=datax[n]-ap; //xi-ap
			diffb=datay[n]-bp; //yi-bp
			diffc=dataz[n]-cp; //cp-zi
			// Li=sqrt( (xi-ap)^2 + (yi-bp)^2 + (zi-cp)^2 )
			Li=sqrt(diffa*diffa + diffb*diffb + diffc*diffc);
			// since dividing by 0 is bad, check to see that Li is > 0
			if(Li>DBL_EPSILON){
                //sum(Li)
				sumLi+=Li;
				//sum((ap-xi)/Li) and similar terms
				sumda+= -diffa/Li;
				sumdb+= -diffb/Li;
				sumdc+= -diffc/Li;
			}else{
				printf(" warning: found Li term equal to 0\n");
			}
		}
		sumLi/=datanum;
		sumda/=datanum;
		sumdb/=datanum;
		sumdc/=datanum;

		// compute radius & a,b,c from sums
		radius=sumLi;
		a=xave + sumLi*sumda;
		b=yave + sumLi*sumdb;
		c=zave + sumLi*sumdc;

		// check to see if approximation is sufficient
		// by testing itteration difference
		diffa=fabs(a-ap);
		diffb=fabs(b-bp);
		diffc=fabs(c-cp);
		if(diffa<=tolerance && diffb<=tolerance && diffc<=tolerance)			
			break;

		//set variables for next itteration
		ap=a,bp=b,cp=c;
		radius=0,a=0,b=0,c=0;
	}

	printf(" Fit sphere with\n");
	printf("    radius:%g\n",radius);
	printf("    origin: a=%g b=%g c=%g\n",a,b,c);
	printf(" after %d iterrations\n",j);
	return (j<maxi); //T=didn't reach max itterations
}


