// sphereMapper.cpp: implementation of the sphereMapper class.

#pragma once
#include "sphereMapper.h"
#include "sphereFit.cpp"
#include "../nodeReader/nodeReader.cpp"

sphereMapper::sphereMapper(int nodes,int nstart,int xres,int yres,double ox,double oy,double oz)
{
	datanum=nodes;
	nodeStart=nstart;
	datax=new double[datanum];
	datay=new double[datanum];
	dataz=new double[datanum];
	pressure=new double[datanum];
	binvalues=new double[xres*yres];
	radius=0.0;	
	a=ox,b=oy,c=oz;
	dx=0.0,dy=0.0;
}

sphereMapper::~sphereMapper()
{
	datanum=0;
	nodeStart=1;
	delete [] datax; datax=NULL;
	delete [] datay; datay=NULL;
	delete [] dataz; dataz=NULL;
	delete [] pressure; pressure=NULL;
	delete [] binvalues; binvalues=NULL;
	radius=0.0;	
	a=0.0,b=0.0,c=0.0;
	dx=0.0,dy=0.0;
}

// LOADNODESK  finds and saves x,y,z coordinates of nodes in fk (.k) file 
//			   in the arrays datax,datay,dataz.
bool sphereMapper::loadNodesK(FILE* fk,FILE* fn)
{
	char line[1000];
	int numFound=0,nodeNumber=1;
	nodeReader nr;
	int nextNode=nodeStart;
	if(fn==NULL){
		nextNode=nodeStart;
	}else{
		nr.setFile(fn);
		nextNode=nr.getNext()+nodeStart-1;
	}

	//find nodes in fk (.k) file
	if(NULL==fgets(line, sizeof(line), fk) ){
		printf("Failed to read first line of fk (.k) file\n");
		return false;
	}
	while(NULL==strstr(line,"*NODE")){
		if(NULL == fgets(line, sizeof(line), fk)){
			printf("Reached end of fk (.k) file before finding nodes\n");
			return false;
		}
	}
	//save nodes in arrays
	while(NULL!=fgets(line,sizeof(line),fk)){
		if(strstr(line,"$"))
		{}
		else if(strstr(line,"*"))
			break;
		else if(1!=sscanf(line,"%d",&nodeNumber)){
			printf("Error reading node number.\n");
			return false;
		}else{
			if(nodeNumber==nextNode){
				if(3!=sscanf(line,"%*d%lf%lf%lf",&datax[numFound],&datay[numFound],&dataz[numFound])){
					printf("Error reading node values.\n");
					return false;
				}
				numFound++;
				if(fn==NULL){
					nextNode++;
				}else{
					nextNode=nr.getNext()+nodeStart-1;
				}
			}
		}
		
	}
	printf(" Found %d nodes in k file\n",numFound);
	if(numFound<datanum){
		printf(" Warning-Only found %d of %d nodes\n",numFound,datanum);
		return false;
	}

	return true;
}

bool sphereMapper::toSphereCoords()
{
	//convert to Sphere coordinates for each data point
	for(int i=0;i<datanum;i++){
		coordToSphere(&datax[i],&datay[i],&dataz[i]);
	}
	return true;
}

bool sphereMapper::toFemurCoords(double ax,double ay,double az,double lx,double ly,double lz,double px,double py,double pz)
{
	//convert given posterior and lateral to sphere coords for exactness.
	coordToSphere(&ax,&ay,&az);
	coordToSphere(&lx,&ly,&lz);
	coordToSphere(&px,&py,&pz);
	printf(" Fit anterior  to sphere: %lf %lf %lf\n",ax,ay,az);
	printf(" Fit lateral   to sphere: %lf %lf %lf\n",lx,ly,lz);
	printf(" Fit posterior to sphere: %lf %lf %lf\n",px,py,pz);
	
	//calculate midpoint of anterior and posterior
	double mx,my,mz;
	mx=(px-ax)/2 +ax;
	my=(py-ay)/2 +ay;
	mz=(pz-az)/2 +az;
	printf(" Calculated midpoint to sphere: %lf %lf %lf\n",mx,my,mz);

	//Determine Vectors of new coord system.
	//y is lateral-origin
	POINT yvector;
	yvector.x=lx-mx,
	yvector.y=ly-my,
	yvector.z=lz-mz;
	//x is posterior-origin
	POINT xvector;
	xvector.x=px-mx,
	xvector.y=py-my,
	xvector.z=pz-mz;
	//z is x cross y
	POINT zvector;
	crossVectors(xvector,yvector,&zvector);
	////y is z cross x (redo for exact coordinate transform
	//crossVectors(zvector,xvector,&yvector);
	//
	//x is y cross z (redo for exact coordinate transform
	crossVectors(yvector,zvector,&xvector);
	
	//divide by vector length to get unit vectors.
	double xvecMag=distance3D(xvector.x,xvector.y,xvector.z,0.0,0.0,0.0);
	xvector.x=xvector.x/xvecMag;
	xvector.y=xvector.y/xvecMag;
	xvector.z=xvector.z/xvecMag;
    double yvecMag=distance3D(yvector.x,yvector.y,yvector.z,0.0,0.0,0.0);
	yvector.x=yvector.x/yvecMag;
	yvector.y=yvector.y/yvecMag;
	yvector.z=yvector.z/yvecMag;
	double zvecMag=distance3D(zvector.x,zvector.y,zvector.z,0.0,0.0,0.0);
	zvector.x=zvector.x/zvecMag;
	zvector.y=zvector.y/zvecMag;
	zvector.z=zvector.z/zvecMag;

	//compute rotation matrix from above vectors to new system
	// these were found by taking the 
	// cross product of each vector with the standard unit
	// vectors (x=[1 0 0],y=[0 1 0], z=[0 0 1]).
	double
		R11=xvector.x,
		R21=xvector.y,
		R31=xvector.z,
		R12=yvector.x,
		R22=yvector.y,
		R32=yvector.z,
		R13=zvector.x,
		R23=zvector.y,
		R33=zvector.z;
	
	//apply rotation to origin to use as offset
	double tempx,tempy,tempz;
	tempx=a;
	tempy=b;
	tempz=c;
	a=R11*tempx+R21*tempy+R31*tempz;
	b=R12*tempx+R22*tempy+R32*tempz;
	c=R13*tempx+R23*tempy+R33*tempz;
	
	//apply transform to every point
	for(int i=0;i<datanum;i++){
		tempx=datax[i];
		tempy=datay[i];
		tempz=dataz[i];
		datax[i]=R11*tempx+R21*tempy+R31*tempz-a;
        datay[i]=R12*tempx+R22*tempy+R32*tempz-b;
		dataz[i]=R13*tempx+R23*tempy+R33*tempz-c-radius;
	}

	//apply transform to lateral to check
	tempx=lx;
	tempy=ly;
	tempz=lz;
	lx=R11*tempx+R21*tempy+R31*tempz-a;
	ly=R12*tempx+R22*tempy+R32*tempz-b;
	lz=R13*tempx+R23*tempy+R33*tempz-c-radius;
	printf(" Fit lateral to femur: %10.6lf %10.6lf %10.6lf\n",lx,ly,lz);
	printf("      should be about: %10.6lf %10.6lf %10.6lf\n",0.0,radius,-radius);
	//apply transform to posterior to check
	tempx=px;
	tempy=py;
	tempz=pz;
	px=R11*tempx+R21*tempy+R31*tempz-a;
	py=R12*tempx+R22*tempy+R32*tempz-b;
	pz=R13*tempx+R23*tempy+R33*tempz-c-radius;
	printf(" Fit posterior to femur: %10.6lf %10.6lf %10.6lf\n",px,py,pz);
	printf("        should be about: %10.6lf %10.6lf %10.6lf\n",radius,0.0,-radius);
	//apply offset to origin to complete transformation
	a-=a;
	b-=b;
	c=c-c-radius;
	printf(" Fit origin to femur: %10.6lf %10.6lf %10.6lf\n",a,b,c);
	printf("          should be : %10.6lf %10.6lf %10.6lf\n",0.0,0.0,-radius);

 	return true;
}

bool sphereMapper::toPlaneCoords()
{
	double length,alpha,d,beta;
	int sign=1;
	double maxX=0.0,minX=0.0,maxY=0.0,minY=0.0;

	//transform to Plane Coords (just x,y) for all data
	for(int i=0;i<datanum;i++){
		length=distance3D(datax[i],datay[i],dataz[i],0.0,0.0,0.0);
		alpha=2*( atan( length/(2*radius) ) );
		d=radius*alpha;
		beta=atan(datay[i]/datax[i]);
		if(datax[i]<0){
			sign=-1;
		}else
			sign=1;
		datax[i]=d*cos(beta)*sign;
		datay[i]=d*sin(beta)*sign;
		dataz[i]=0.0;
		if(fabs(datax[i])>maxX) maxX=fabs(datax[i]);
		if(fabs(datay[i])>maxY) maxY=fabs(datay[i]);
	}
	//transform center to plane coordinates 
	// with slight offset so it won't overlap the shells
	a=0.0,b=0.0,c=-2.0;

	dx=maxX*2;
	dy=maxY*2;

	return true;
}

bool sphereMapper::loadPressures(FILE* fp,FILE* fn,int pressureStart)
{
	char line[1000];
	int nodeNumber=1;
	nodeReader nr;
	int nextNode=0;
	if(fn==NULL){
		nextNode=pressureStart;
	}else{
		nr.setFile(fn);
		nextNode=nr.getNext()+pressureStart-1;
	}

	
	//find NODAL_RESULTS in fin (pressure) file
	if(NULL==fgets(line, sizeof(line), fp) ){
		printf("Failed to read first line of fin (pressure) file\n");
		return false;
	}
	while(NULL==strstr(line,"*NODAL_RESULTS")){
		if(NULL == fgets(line, sizeof(line), fp)){
			printf("Reached end of fp (pressure) file before finding pressures\n");
			return false;
		}
	}
	//save pressures in array
	int i=0;
	for(i=0;i<datanum;i++){
		if(NULL==fgets(line,sizeof(line),fp)){
			printf("Couldn't read pressure for node # %d\n",i);
			return false;
		}
		
		if(strstr(line,"$")){
			i--;
		}else if(strstr(line,"*")){
			i--;
			break;
		}else if(1!=sscanf(line,"%d",&nodeNumber)){
			printf("Error reading node number.\n");
			return false;
		}else{
			if(nodeNumber==nextNode){
				if(1!=sscanf(line,"%*d%lf",&pressure[i])){
					printf("Error reading pressure value.\n");
					return false;
				}
				if(pressure[i]<=0.000001){
					pressure[i]=0.000001;
				}
				if(fn==NULL){
					nextNode++;
				}else{
					nextNode=nr.getNext()+pressureStart-1;
				}
			}else{
				i--;
			}
		}
	}
	printf(" Found pressures for %d nodes.\n",i);
	if(i!=datanum){
		printf("Failed to find all %d pressures\n",datanum);
		return false;
	}

	return true;
}

bool sphereMapper::toImageBins(int xres,int yres)
{
	double bwx=dx/xres,bwy=dy/yres;
	int binnum=0;
	double offsetx=(bwx-dx)/2,offsety=(bwy-dy)/2;  //for center of bin
	POINT binCenter;
	binCenter.z=0.0;
	int nn1=0,nn2=0,nn3=0;
	double fromC=0;

	for(int j=0;j<yres;j++){
		binCenter.y=offsety+(bwy*j);
		for(int i=0;i<xres;i++){
			binCenter.x=offsetx+(bwx*i);
			binnum=i+j*xres;
			
			//test for in mesh area
			fromC=distance3D(binCenter.x,binCenter.y,binCenter.z,0.0,0.0,0.0);
			if(fromC<=dx/2||fromC<=dy/2){
				//find nearest nodes
				nearestNodes(binCenter,&nn1,&nn2,&nn3);
				//assign binvalue to average of node pressures
				binvalues[binnum]=(pressure[nn1]+pressure[nn2]+pressure[nn3])/3;
				if(binvalues[binnum]<0.000001){
					binvalues[binnum]=0.000001;
				}
			}else{
				binvalues[binnum]=500.0;
			}
			binnum++;
			}
	}

	return true;
}

bool sphereMapper::writeImage(FILE* fout,FILE* fe,int xres,int yres,int option)
{
	switch (option)
	{
	case 1:
		printf(" Printing Nodes with current coordinates (appending)...\n");
		if(!this->printSphereNodes(fout)){
				printf(" Failed to print sphere nodes\n");
				return false;
		}
		break;
	case 2:
		printf(" Printing Elements of Sphere (appending)...\n");
		if(!this->printSphereElements(fout,fe)){
			printf(" Failed to print sphere elements\n");
			return false;
		}
		break;
	case 3:
		printf(" Printing Nodes and Element Shells to 2D circle with current pressures (appending)...\n");
		if(!this->printCircle(fout,xres,yres)){
			printf(" Failed to print circle to file\n");
			return false;
		}
		break;
	default:
		printf(" Invalid option given for write image\n");
		return false;
	}
	
	return true;
}


bool sphereMapper::printSphereNodes(FILE* fout)
{
	fprintf(fout,"$Sphere Nodes\n");
	fprintf(fout,"*NODE\n");
	for(int i=0;i<datanum;i++){
		fprintf(fout,"%8d%16.8lf%16.8lf%16.8lf\n",i+1,datax[i],datay[i],dataz[i]);
	}

	//Check to print center node
	int toPrint=0;
	printf("\nWould you like to print the center node in addition? (yes=1 no=0)\n");
	scanf("%d",&toPrint);
	if(toPrint==1){
		fprintf(fout,"%8d%16.8lf%16.8lf%16.8lf\n",datanum+1,a,b,c);
	}

	return true;
}
bool sphereMapper::printSphereElements(FILE* fout, FILE* fe)
{
	char line[1000];
	
	//find elements in fe (.k) file
	if(NULL==fgets(line, sizeof(line), fe) ){
		printf("Failed to read first line of element file (fe.k)\n");
		return false;
	}
	while(NULL==strstr(line,"*ELEMENT_SOLID")){
		if(NULL == fgets(line, sizeof(line), fe)){
			printf("Reached end of element file before finding *ELEMENT_SOLID\n");
			return false;
		}
	}
	
	//Scan and Print elements
	int numFound=0;

	fprintf(fout,"$Sphere Elements\n");
	fprintf(fout,"*ELEMENT_SOLID\n");
	while(NULL!=fgets(line,sizeof(line),fe)){
		if(strstr(line,"$"))
		{}
		else if(strstr(line,"*"))
			break;
		else{
			//print element line
			fprintf(fout,"%s",line);
			numFound++;
		}
	}
		
	printf(" Printed %d elements for sphere\n",numFound);
	return true;
}
bool sphereMapper::printCircle(FILE* fout,int xres,int yres)
{
	double bwx=dx/xres,bwy=dy/yres;
	int binnum=0,nodenum=datanum+1;
	double offsetx=(bwx-dx)/2,offsety=(bwy-dy)/2;  //for center of bin
	double binCenterx=0.0,binCentery=0.0,binCenterz=0.0;

	//print nodes
	fprintf(fout,"$Grid Nodes\n");
	fprintf(fout,"*NODE\n");
	for(int j=0;j<yres;j++){
		binCentery=offsety+(bwy*j);
		for(int i=0;i<xres;i++){
			binCenterx=offsetx+(bwx*i);
			binnum=i+j*xres;
			fprintf(fout,"%8d%16.8lf%16.8lf%16.8lf\n",nodenum++,binCenterx,binCentery,binCenterz);
		}
	}

    //print connectivity
	int corner=0;
	int c1=0,c2=0,c3=0,c4=0;
	binnum=0,nodenum=datanum;

	fprintf(fout,"$Grid Connectivity with Pressure Shells\n");
	fprintf(fout,"*ELEMENT_SHELL_THICKNESS\n");
	for(int n=0;n<yres-1;n++){
		for(int m=0;m<xres-1;m++){
			c1=corner+1;
			c2=corner+2;
			c3=corner+xres+2;
			c4=corner+xres+1;
			if(binvalues[c1-1]==500.0||binvalues[c2-1]==500.0||binvalues[c3-1]==500.0||binvalues[c4-1]==500.0)
				fprintf(fout,"%8d%8d%8d%8d%8d%8d\n",binnum+1,1,c1+nodenum,c2+nodenum,c3+nodenum,c4+nodenum);
			else
				fprintf(fout,"%8d%8d%8d%8d%8d%8d\n",binnum+1,2,c1+nodenum,c2+nodenum,c3+nodenum,c4+nodenum);
			fprintf(fout,"%16.8f%16.8f%16.8f%16.8f\n",binvalues[c1-1],binvalues[c2-1],binvalues[c3-1],binvalues[c4-1]);
			binnum++;
			corner++;
		}
		corner++;
	}

	return true;
}

double sphereMapper::distance3D(double x1, double y1, double z1, double x2, double y2, double z2)	
{
	double d=(x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2);
	d=sqrt(d);
	return d;
}

bool sphereMapper::coordToSphere(double* x, double* y, double* z)
{
	double length=0;

	length=distance3D(*x,*y,*z,a,b,c);
	
	//uses similar triangle ratio to get closest point on sphere
	*x=((*x-a)*radius/length)+a;
	*y=((*y-b)*radius/length)+b;
	*z=((*z-c)*radius/length)+c;

	return true;
}

bool sphereMapper::crossVectors(POINT first, POINT second, POINT* output)
{
	//x=fy*sz - fz*sy
	(*output).x=first.y*second.z - first.z*second.y;
	//y=fz*sx - fx*sz
	(*output).y=first.z*second.x - first.x*second.z;
	//z=fx*sy - fy*sx
	(*output).z=first.x*second.y - first.y*second.x;

	return true;
}

bool sphereMapper::nearestNodes(POINT node, int* out1, int* out2, int* out3)
{
	int num1=0,num2=0,num3=0;
	double testdist=0,dist1=radius,dist2=radius,dist3=radius;
	
    for(int i=0;i<datanum;i++){
		testdist=distance3D(datax[i],datay[i],dataz[i],node.x,node.y,node.z);
		if(testdist<dist1){
			dist3=dist2;
			num3=num2;
			dist2=dist1;
			num2=num1;
			dist1=testdist;
			num1=i;
		}else if(testdist<dist2){
			dist3=dist2;
			num3=num2;
			dist2=testdist;
			num2=i;
		}else if(testdist<dist3){
			dist3=testdist;
			num3=i;
		}
	}

	*out1=num1;
	*out2=num2;
	*out3=num3;

	return true;
}




	

